from dataclasses import dataclass
from typing import Optional

from rich import print
from rich.panel import Panel
from rich.console import RenderGroup

from recipys.types import Printable
from recipys.constants import (
    COLOR_RECIPE_MAIN,
    COLOR_RECIPE_SECONDARY,
    COLOR_ERROR,
    COLOR_WARNING,
    COLOR_INFORMATION,
)


@dataclass
class ConsolePrinter:
    """Prints recipe information to console"""

    recipe: Printable

    def print_recipe(self) -> None:
        """Print recipe or error message"""
        if self.recipe.error_message:
            self._print_error()
        elif self.recipe.warning_message:
            self._print_warning()
        elif self.recipe.info_message:
            self._print_information()
        else:
            self._print_recipe_info()

    def _print_message(
        self,
        title: str,
        message: str,
        color: str = "white",
        subtitle: Optional[str] = None,
        render_group: Optional[RenderGroup] = None,
    ) -> None:
        """Generic print message"""
        if not render_group:
            render_group = RenderGroup(
                f"[{color}]{subtitle}",
                message,
            )
        panel = Panel.fit(
            render_group,
            title=f"[bold {color}]" + f"  {title}  ",
            border_style=f"bold {color}",
            padding=(1, 1),
            title_align="left",
        )
        print("\n")
        print(panel)
        print("\n")

    def _make_inner_panel(self, text: str, title: str) -> Panel:
        """Create interior panel for printed recipes"""
        return Panel(
            text,
            title=f"[bold {COLOR_RECIPE_SECONDARY}]" + title,
            padding=(1, 1),
            border_style=f"bold {COLOR_RECIPE_SECONDARY}",
            title_align="left",
        )

    def _print_recipe_info(self) -> None:
        """Print recipe info"""
        render_group = RenderGroup(
            self._make_inner_panel(self.recipe.ingredients, "Ingredients"),
            self._make_inner_panel(self.recipe.preparation, "Preparation"),
        )

        return self._print_message(
            title=self.recipe.title,
            message=self.recipe.warning_message,
            color=COLOR_RECIPE_MAIN,
            render_group=render_group,
        )

    def _print_error(self) -> None:
        """Print error message"""
        return self._print_message(
            title="ERROR",
            subtitle="The following error ocurred:",
            message=self.recipe.error_message,
            color=COLOR_ERROR,
        )

    def _print_warning(self):
        """Print warning message"""
        return self._print_message(
            title="Warning",
            subtitle="Heads up!",
            message=self.recipe.warning_message,
            color=COLOR_WARNING,
        )

    def _print_information(self):
        """Print information message"""
        return self._print_message(
            title=self.recipe.title,
            subtitle="Your requested info:",
            message=self.recipe.info_message,
            color=COLOR_INFORMATION,
        )



import sys
from recipys.ArgParser import ArgParser
from recipys.ConsolePrinter import ConsolePrinter
from recipys.RecipeFetcher import RecipeFetcher
from recipys.request_wait import wait_for_green_light
from recipys.types import PrintInterrupt


def main():
    """Main entry point of app"""

    wait_for_green_light()

    try:
        recipe_constraints = ArgParser(sys.argv).parse()
    except PrintInterrupt as e:
        recipe = e.printable
    else:
        recipe = RecipeFetcher(recipe_constraints).fetch()
    finally:
        ConsolePrinter(recipe).print_recipe()


if __name__ == "__main__":
    main()  # pragma: no cover



from dataclasses import dataclass
import re
from time import sleep
import random

from recipys.types import RecipeConstraints, Printable, FetchingError
from recipys.Scraper import (
    Scraper,
    ScraperSearchTerms,
    HtmlSearchTarget,
)
from recipys.constants import WAIT_BETWEEN_PINGS, KEY_STRINGS_CUT_FROM_RECIPE


@dataclass
class RecipeFetcher:
    """Fetches recipes from the web"""

    recipe_constraints: RecipeConstraints

    def fetch(self) -> Printable:
        """Fetches recipe according to user input"""
        try:
            self.recipe_url = self._scrape_recipe_url()
            sleep(WAIT_BETWEEN_PINGS)
            recipe = self._scrape_recipe()
            return recipe
        except FetchingError as e:
            return Printable(error_message=e.message)

    def _create_url_recipe_search(self) -> str:
        """Get URL for HTTP GET request"""
        url_base: str = "https://www.recipe-free.com/recipe/"

        if not hasattr(self, "_target_page_recipe_list"):
            self._target_page_recipe_list: str = "1"
        url_suffix: str = f"/{self._target_page_recipe_list}/search"

        ings = (
            self.recipe_constraints.ingredients
            if self.recipe_constraints.ingredients
            else [None]
        )
        query = [self.recipe_constraints.meal] + ings
        clean_query = [e for e in query if e]

        if clean_query:
            return url_base + "-".join(clean_query) + url_suffix
        else:
            return (
                "https://www.recipe-free.com/recipe/best/"
                f"{self._target_page_recipe_list}"
            )

    def _create_search_terms_recipe_search(self) -> None:
        """Creates search terms used to search for recipe url"""
        if not hasattr(self, "terms_recipe_list"):
            self.terms_recipe_list = ScraperSearchTerms(
                target=HtmlSearchTarget(
                    name="Recipe",
                    tag="a",
                    att_name="class",
                    att_value="day",
                    target_element="href",
                ),
                return_multiple=True,
            )
        if not hasattr(self, "terms_pages_with_recipes"):
            self.terms_pages_with_recipes = ScraperSearchTerms(
                target=HtmlSearchTarget(
                    name="Pages",
                    tag="span",
                    att_name="class",
                    att_value="f12 f12",
                )
            )

    def _scrape_recipe_url(self) -> str:
        """Scrape url of recipe according to search constraints"""
        self._create_search_terms_recipe_search()
        self._url_recipe_list = self._create_url_recipe_search()

        scraper = Scraper(
            url=self._url_recipe_list,
            search_terms=[
                self.terms_recipe_list,
                self.terms_pages_with_recipes,
            ],
        )
        scraped_data = scraper.get()

        try:
            pages_text = scraped_data.get("Pages")[0]
            if not self._is_current_recipe_page_the_right_one(pages_text):
                sleep(WAIT_BETWEEN_PINGS)
                return self._scrape_recipe_url()

            recipes_urls = scraped_data.get("Recipe")
            return random.choice(recipes_urls)

        except IndexError:
            raise FetchingError(
                "No recipe found with your criteria. "
                "Maybe try removing or changing your filters "
                "to broaden your search? Using only ingredients could help!"
            )

    def _scrape_recipe(self) -> Printable:
        """Scrape recipe information from its URL"""
        scraper = self._setup_scraper_recipe()
        results = scraper.get()

        try:
            recipe_title = self._beautify(results["Title"][0])
            recipe_ingredients = self._beautify(
                results["Ingredients & Preparation"][0]
            )
            recipe_preparation = self._beautify(
                results["Ingredients & Preparation"][1]
            )
        except KeyError or IndexError:
            raise FetchingError("Recipe format incorrect. Please try again")
        else:
            return Printable(
                title=recipe_title,
                ingredients=recipe_ingredients,
                preparation=recipe_preparation,
            )

    def _setup_scraper_recipe(self) -> Scraper:
        """Setup Scraper object to be used scraping recipe from its page"""
        target_title = HtmlSearchTarget(
            name="Title",
            tag="h1",
            att_name="class",
            att_value="red",
        )
        target_ings_and_prep = HtmlSearchTarget(
            name="Ingredients & Preparation",
            tag="p",
            att_name="style",
            att_value="padding-left: 30px",
        )

        search_terms_title = ScraperSearchTerms(target=target_title)
        search_terms_ings_and_prep = ScraperSearchTerms(
            target=target_ings_and_prep, return_multiple=True
        )

        scraper = Scraper(
            url=self.recipe_url,
            search_terms=[
                search_terms_title,
                search_terms_ings_and_prep,
            ],
        )

        return scraper

    def _beautify(self, text: str) -> str:
        """Return beautified text from parsed html

        Args:
            - text: text to be beautified
        """

        # Remove metadata if key string found
        pos_to_cut: int = 0
        text_lowercase = text.lower()
        for key in KEY_STRINGS_CUT_FROM_RECIPE:
            substring_position = text_lowercase.find(key)
            if substring_position != -1:
                # Substring found
                if pos_to_cut == 0:
                    pos_to_cut = substring_position
                else:
                    pos_to_cut = min(pos_to_cut, substring_position)
        text_without_metadata = text[:pos_to_cut] if pos_to_cut else text

        # Remove multiple spaces and leading and trailing spaces
        return re.sub(" +", " ", text_without_metadata).strip()

    def _is_current_recipe_page_the_right_one(self, text: str) -> bool:
        """Determine if we are in the right page of recipe listing

        Args:
            - text: Inner HTML text of total pages information
        """
        if hasattr(self, "_determined_target_page"):
            return True

        regex = re.compile(r"Results (\d).+?\(de (.*?)\) recipes")
        m = regex.match(text)

        recipes_per_page: int = 10

        current_page = (int(m.group(1)) // recipes_per_page) + 1

        recipes_total = int(m.group(2).replace(",", ""))
        pages_total: int = (recipes_total // recipes_per_page) + 1

        target_page: int = random.randint(1, pages_total)
        self._target_page_recipe_list = str(target_page)

        self._determined_target_page: bool = True

        return target_page == current_page



"""
Global constants of application
"""

# Real time delay between two executions of application (ethical scraping)
from typing import List


MIN_SECONDS_REQUEST: float = 10.0


# Seconds to wait between requests to same endpoint (ethical scraping)
WAIT_BETWEEN_PINGS: int = 1


# Rich colors for printed messages
COLOR_RECIPE_MAIN: str = "steel_blue"
COLOR_RECIPE_SECONDARY: str = "light_slate_grey"
COLOR_ERROR: str = "orange_red1"
COLOR_WARNING: str = "gold1"
COLOR_INFORMATION: str = "deep_sky_blue1"


# Standard messages
MESSAGE_INVALID_ARGS: str = f"""\
It seems the input arguments are invalid

You can only use [bold {COLOR_ERROR}]one[/bold {COLOR_ERROR}] \
of the accepted meals :
[bold {COLOR_ERROR}]breakfast[/bold {COLOR_ERROR}], \
[bold {COLOR_ERROR}]lunch[/bold {COLOR_ERROR}], \
[bold {COLOR_ERROR}]dinner[/bold {COLOR_ERROR}] or \
[bold {COLOR_ERROR}]desert[/bold {COLOR_ERROR}]

Or if you are specifying ingredients, be sure to use the keyword \
[bold {COLOR_INFORMATION}]with[/bold {COLOR_INFORMATION}] before \
the ingredients list (for example: [bold {COLOR_INFORMATION}]\
recipys with potato beans[/bold {COLOR_INFORMATION}])
"""

MESSAGE_INVALID_INGREDIENT: str = f"""\
It seems the input arguments are invalid

Please make sure to specify valid ingredients after the \
[bold {COLOR_ERROR}]with[/bold {COLOR_ERROR}]' keyword
(valid ingredients have at least one letter, not only symbols)
"""


# Key strings to cut recipe metadata, as starting point
KEY_STRINGS_CUT_FROM_RECIPE: List[str] = [
    "source: ",
    "submitted by",
    "from: ",
    "recipe by",
]



from dataclasses import dataclass
import re
from typing import List, Tuple, Optional

from recipys.types import RecipeConstraints, PrintInterrupt, Printable
from recipys.constants import MESSAGE_INVALID_ARGS, MESSAGE_INVALID_INGREDIENT


@dataclass
class ArgParser:
    """Command line arguments parser"""

    args: List[str]
    accepted_meals: Tuple[str] = (
        "breakfast",
        "lunch",
        "dinner",
        "dessert",
    )

    def parse(self) -> RecipeConstraints:
        """Parses command line arguments"""
        self.meal = self._get_meal()
        self.ingredients = self._get_ingredients()
        return RecipeConstraints(self.meal, self.ingredients)

    def _get_meal(self) -> Optional[str]:
        """Gets meal type from command line argument

        Raises:
            - PrintInterrupt: If error detected in parsed meal
        """
        if len(self.args) > 1:
            for accepted_meal in self.accepted_meals:
                if self.args[1].lower() == accepted_meal:
                    return accepted_meal

            # No meal found: check if error
            if "with" != self.args[1]:
                raise PrintInterrupt(
                    Printable(error_message=MESSAGE_INVALID_ARGS)
                )
        return None

    def _get_ingredients(self) -> Optional[List[str]]:
        """Gets ingredients from command line argument

        Raises:
            - PrintInterrupt: If error detected in parsed ingredients"""
        args = [a.lower() for a in self.args]
        start: int = 2 if self.meal else 1

        if len(args) >= start + 1:
            if "with" != args[start]:
                raise PrintInterrupt(
                    Printable(error_message=MESSAGE_INVALID_ARGS)
                )

        ingredients: List[str] = []
        for _, arg in enumerate(args[start + 1 : :]):
            m = re.findall(r"[a-zA-Z]", arg)
            word = "".join(m)
            if word:
                ingredients.append(word)
            else:
                raise PrintInterrupt(
                    Printable(error_message=MESSAGE_INVALID_INGREDIENT)
                )

        if not ingredients:
            return None

        return ingredients
